@@ -75,6 +75,9 @@ struct ECHState {
   Random random;
   // ECH handshake context (initialized during HRR)
   mutable std::unique_ptr<HandshakeContext> handshakeContext;
+  // In the case of rejection, the server may send a list of supported
+  // configs, which we store here.
+  folly::Optional<std::vector<ech::ECHConfig>> retryConfigs;
 };
 
 class State {
