@@ -178,7 +178,7 @@ class MockPeerCert : public PeerCert {
 class MockCertificateVerifier : public CertificateVerifier {
  public:
   MOCK_METHOD(
-      void,
+      std::shared_ptr<const folly::AsyncTransportCertificate>,
       verify,
       (const std::vector<std::shared_ptr<const PeerCert>>&),
       (const));
@@ -224,7 +224,12 @@ class MockFactory : public OpenSSLFactory {
       (const));
   MOCK_METHOD(
       std::unique_ptr<KeyExchange>,
-      makeKeyExchange,
+      makeClientKeyExchange,
+      (NamedGroup group),
+      (const));
+  MOCK_METHOD(
+      std::unique_ptr<KeyExchange>,
+      makeServerKeyExchange,
       (NamedGroup group),
       (const));
   MOCK_METHOD(std::unique_ptr<Aead>, makeAead, (CipherSuite cipher), (const));
@@ -279,11 +284,18 @@ class MockFactory : public OpenSSLFactory {
           ret->setDefaults();
           return ret;
         }));
-    ON_CALL(*this, makeKeyExchange(_)).WillByDefault(InvokeWithoutArgs([]() {
-      auto ret = std::make_unique<NiceMock<MockKeyExchange>>();
-      ret->setDefaults();
-      return ret;
-    }));
+    ON_CALL(*this, makeClientKeyExchange(_))
+        .WillByDefault(InvokeWithoutArgs([]() {
+          auto ret = std::make_unique<NiceMock<MockKeyExchange>>();
+          ret->setDefaults();
+          return ret;
+        }));
+    ON_CALL(*this, makeServerKeyExchange(_))
+        .WillByDefault(InvokeWithoutArgs([]() {
+          auto ret = std::make_unique<NiceMock<MockKeyExchange>>();
+          ret->setDefaults();
+          return ret;
+        }));
     ON_CALL(*this, makeAead(_)).WillByDefault(InvokeWithoutArgs([]() {
       auto ret = std::make_unique<NiceMock<MockAead>>();
       ret->setDefaults();
@@ -375,6 +387,7 @@ class MockAsyncFizzBase : public AsyncFizzBase {
 
   MOCK_METHOD(folly::Optional<Random>, getClientRandom, (), (const));
   MOCK_METHOD(void, tlsShutdown, ());
+  MOCK_METHOD(void, initiateKeyUpdate, (KeyUpdateRequest), (override));
 
   MOCK_METHOD(
       void,
