@@ -111,6 +111,12 @@ FIZZ_DECLARE_EVENT_HANDLER(
     Event::AppWrite,
     StateEnum::Error);
 
+FIZZ_DECLARE_EVENT_HANDLER(
+    ServerTypes,
+    StateEnum::AcceptingData,
+    Event::KeyUpdateInitiation,
+    StateEnum::AcceptingData);
+
 FIZZ_DECLARE_EVENT_HANDLER(
     ServerTypes,
     StateEnum::AcceptingData,
@@ -205,6 +211,12 @@ Actions ServerStateMachine::processAppCloseImmediate(const State& state) {
   return detail::handleAppCloseImmediate(state);
 }
 
+AsyncActions ServerStateMachine::processKeyUpdateInitiation(
+    const State& state,
+    KeyUpdateInitiation keyUpdateInitiation) {
+  return detail::processEvent(state, std::move(keyUpdateInitiation));
+}
+
 namespace detail {
 
 AsyncActions processEvent(const State& state, Param param) {
@@ -436,8 +448,6 @@ static Optional<ProtocolVersion> negotiateVersion(
 
 static Optional<CookieState> getCookieState(
     const ClientHello& chlo,
-    ProtocolVersion version,
-    CipherSuite cipher,
     const CookieCipher* cookieCipher) {
   auto cookieExt = getExtension<Cookie>(chlo.extensions);
   if (!cookieExt) {
@@ -458,6 +468,22 @@ static Optional<CookieState> getCookieState(
         "could not decrypt cookie", AlertDescription::decrypt_error);
   }
 
+  if (cookieState->echCipherSuite.has_value() ^
+      (cookieState->echConfigId != nullptr)) {
+    throw FizzException(
+        "cookie has incomplete ech params", AlertDescription::internal_error);
+  }
+
+  return cookieState;
+}
+
+static void verifyCookieState(
+    Optional<CookieState>& cookieState,
+    ProtocolVersion version,
+    CipherSuite cipher) {
+  if (!cookieState) {
+    return;
+  }
   if (cookieState->version != version) {
     throw FizzException(
         "version mismatch with cookie", AlertDescription::protocol_version);
@@ -467,8 +493,6 @@ static Optional<CookieState> getCookieState(
     throw FizzException(
         "cipher mismatch with cookie", AlertDescription::handshake_failure);
   }
-
-  return cookieState;
 }
 
 namespace {
@@ -589,11 +613,6 @@ static std::
   auto scheduler = factory.makeKeyScheduler(cipher);
 
   if (cookieState) {
-    if (handshakeContext) {
-      throw FizzException(
-          "cookie after statefull hrr", AlertDescription::illegal_parameter);
-    }
-
     handshakeContext = factory.makeHandshakeContext(cipher);
 
     message_hash chloHash;
@@ -693,7 +712,7 @@ static Buf doKex(
     NamedGroup group,
     const Buf& clientShare,
     KeyScheduler& scheduler) {
-  auto kex = factory.makeKeyExchange(group);
+  auto kex = factory.makeServerKeyExchange(group);
   kex->generateKeyPair();
   auto sharedSecret = kex->generateSharedSecret(clientShare->coalesce());
   scheduler.deriveHandshakeSecret(sharedSecret->coalesce());
@@ -891,6 +910,7 @@ static Buf getEncryptedExt(
     HandshakeContext& handshakeContext,
     const folly::Optional<std::string>& selectedAlpn,
     EarlyDataType earlyData,
+    folly::Optional<std::vector<ech::ECHConfig>> echRetryConfigs,
     std::vector<Extension> otherExtensions) {
   EncryptedExtensions encryptedExt;
   if (selectedAlpn) {
@@ -905,6 +925,12 @@ static Buf getEncryptedExt(
     encryptedExt.extensions.push_back(encodeExtension(ServerEarlyData()));
   }
 
+  if (echRetryConfigs.has_value()) {
+    ech::ServerECH serverEch;
+    serverEch.retry_configs = std::move(*echRetryConfigs);
+    encryptedExt.extensions.push_back(encodeExtension(std::move(serverEch)));
+  }
+
   for (auto& ext : otherExtensions) {
     encryptedExt.extensions.push_back(std::move(ext));
   }
@@ -995,25 +1021,130 @@ static Buf getCertificateRequest(
   return encodedCertificateRequest;
 }
 
+static ECHStatus processECHHRR(
+    const Optional<CookieState>& cookieState,
+    const State& state,
+    ClientHello& chlo) {
+  auto decrypter = state.context()->getECHDecrypter();
+  auto echExt = getExtension<ech::ClientECH>(chlo.extensions);
+  ECHStatus echStatus = state.echStatus();
+
+  // Check for cookie ECH
+  bool cookieHasECH = cookieState && cookieState->echCipherSuite.has_value();
+
+  // Do checks for both cases.
+  if (echStatus == ECHStatus::Accepted || cookieHasECH) {
+    if (!decrypter) {
+      throw FizzException(
+          "ech accepted but no decrypter set up",
+          AlertDescription::internal_error);
+    }
+    if (!echExt) {
+      throw FizzException(
+          "ech not sent for hrr", AlertDescription::missing_extension);
+    }
+    if (!echExt->config_id->empty() || !echExt->enc->empty()) {
+      throw FizzException(
+          "hrr ech enc or config_id not empty",
+          AlertDescription::illegal_parameter);
+    }
+  }
+
+  if (echStatus == ECHStatus::Accepted) {
+    // Stateful HRR first.
+    if (state.echState()->cipherSuite != echExt->cipher_suite) {
+      throw FizzException(
+          "ech hrr cipher suite mismatch", AlertDescription::illegal_parameter);
+    }
+
+    chlo = decrypter->decryptClientHelloHRR(
+        chlo, state.echState()->configId, state.echState()->hpkeContext);
+
+    return ECHStatus::Accepted;
+  } else if (cookieHasECH) {
+    // Stateless HRR now
+    if (*cookieState->echCipherSuite != echExt->cipher_suite) {
+      throw FizzException(
+          "ech hrr cipher suite mismatch", AlertDescription::illegal_parameter);
+    }
+
+    chlo = decrypter->decryptClientHelloHRR(
+        chlo, cookieState->echConfigId, cookieState->echEnc);
+
+    return ECHStatus::Accepted;
+  }
+
+  // Just return the ECH status as is
+  return echStatus;
+}
+
+// Process ECH, replacing chlo if successful
+static std::pair<ECHStatus, ECHState> processECH(
+    const Optional<CookieState>& cookieState,
+    const State& state,
+    ClientHello& chlo) {
+  // First, fetch current state (if any).
+  ECHState echState;
+  ECHStatus echStatus = state.echStatus();
+  auto decrypter = state.context()->getECHDecrypter();
+
+  if (state.handshakeContext() || cookieState) {
+    // Process ECH for HRR (if any)
+    echStatus = processECHHRR(cookieState, state, chlo);
+    // Populate ECH state for saving
+    if (state.echState().has_value()) {
+      echState.hpkeContext = std::move(state.echState()->hpkeContext);
+      if (state.echState()->configId != nullptr) {
+        echState.configId = state.echState()->configId->clone();
+      }
+      echState.cipherSuite = state.echState()->cipherSuite;
+    }
+  } else {
+    bool requestedECH =
+        findExtension(chlo.extensions, ExtensionType::encrypted_client_hello) !=
+        chlo.extensions.end();
+    if (requestedECH && decrypter) {
+      auto gotChlo = decrypter->decryptClientHello(chlo);
+      if (gotChlo.has_value()) {
+        auto echExt = getExtension<ech::ClientECH>(chlo.extensions);
+        echStatus = ECHStatus::Accepted;
+        echState.hpkeContext = std::move(gotChlo->context);
+        echState.configId = std::move(gotChlo->configId);
+        echState.cipherSuite = echExt->cipher_suite;
+        chlo = std::move(gotChlo->chlo);
+      } else if (requestedECH) {
+        echStatus = ECHStatus::Rejected;
+      }
+    }
+  }
+
+  // Check for ECH inner extension, if accepted.
+  auto innerExt = getExtension<ech::ECHIsInner>(chlo.extensions);
+  if (echStatus == ECHStatus::Accepted && !innerExt) {
+    throw FizzException(
+        "inner clienthello missing ech_is_inner",
+        AlertDescription::missing_extension);
+  }
+
+  return std::make_pair(echStatus, std::move(echState));
+}
+
 AsyncActions
 EventHandler<ServerTypes, StateEnum::ExpectingClientHello, Event::ClientHello>::
     handle(const State& state, Param param) {
   ClientHello chlo = std::move(*param.asClientHello());
 
-  // Update the client hello if we are using ECH
-  auto decrypter = state.context()->getECHDecrypter();
-  bool acceptedECH = false;
-  bool requestedECH =
-      findExtension(chlo.extensions, ExtensionType::encrypted_client_hello) !=
-      chlo.extensions.end();
-  if (decrypter) {
-    auto gotChlo = decrypter->decryptClientHello(chlo);
-    if (gotChlo.has_value()) {
-      acceptedECH = true;
-      chlo = std::move(gotChlo->chlo);
-    }
+  auto cookieState = getCookieState(chlo, state.context()->getCookieCipher());
+
+  if (state.handshakeContext() && cookieState.has_value()) {
+    throw FizzException(
+        "cookie after statefull hrr", AlertDescription::illegal_parameter);
   }
 
+  ECHStatus echStatus;
+  ECHState echState;
+  std::tie(echStatus, echState) = processECH(cookieState, state, chlo);
+
   addHandshakeLogging(state, chlo);
 
   if (state.readRecordLayer()->hasUnparsedHandshakeData()) {
@@ -1060,8 +1191,7 @@ EventHandler<ServerTypes, StateEnum::ExpectingClientHello, Event::ClientHello>::
 
   auto cipher = negotiateCipher(chlo, state.context()->getSupportedCiphers());
 
-  auto cookieState = getCookieState(
-      chlo, *version, cipher, state.context()->getCookieCipher());
+  verifyCookieState(cookieState, *version, cipher);
 
   auto resStateResult = getResumptionState(
       chlo,
@@ -1088,8 +1218,8 @@ EventHandler<ServerTypes, StateEnum::ExpectingClientHello, Event::ClientHello>::
        version = *version,
        cipher,
        pskMode = resStateResult.pskMode,
-       acceptedECH,
-       requestedECH,
+       echStatus,
+       echState = std::move(echState),
        obfuscatedAge =
            resStateResult.obfuscatedAge](FutureResultType result) mutable {
         auto& resumption = *std::get<0>(result);
@@ -1252,8 +1382,10 @@ EventHandler<ServerTypes, StateEnum::ExpectingClientHello, Event::ClientHello>::
                              group,
                              earlyDataType,
                              replayCacheResult,
-                             newReadRecordLayer = std::move(
-                                 newReadRecordLayer)](State& newState) mutable {
+                             newReadRecordLayer = std::move(newReadRecordLayer),
+                             echStatus,
+                             echState =
+                                 std::move(echState)](State& newState) mutable {
                   // Save some information about the current state to be
                   // validated when we get the second client hello. We don't
                   // validate that the second client hello matches the first
@@ -1267,6 +1399,10 @@ EventHandler<ServerTypes, StateEnum::ExpectingClientHello, Event::ClientHello>::
                   newState.earlyDataType() = earlyDataType;
                   newState.replayCacheResult() = replayCacheResult;
                   newState.readRecordLayer() = std::move(newReadRecordLayer);
+                  newState.echStatus() = echStatus;
+                  if (echStatus == ECHStatus::Accepted) {
+                    newState.echState() = std::move(echState);
+                  }
                 }),
                 std::move(serverFlight),
                 MutateState(&Transition<StateEnum::ExpectingClientHello>)));
@@ -1315,11 +1451,15 @@ EventHandler<ServerTypes, StateEnum::ExpectingClientHello, Event::ClientHello>::
             std::move(serverShare),
             legacySessionId ? legacySessionId->clone() : nullptr);
 
-        if (acceptedECH) {
+        folly::Optional<std::vector<ech::ECHConfig>> echRetryConfigs;
+        if (echStatus == ECHStatus::Accepted) {
           // If accepted, go ahead and set the last random bytes to the
           // accept_confirmation
           ech::setAcceptConfirmation(
               serverHello, handshakeContext->clone(), scheduler);
+        } else if (echStatus == ECHStatus::Rejected) {
+          auto decrypter = state.context()->getECHDecrypter();
+          echRetryConfigs = decrypter->getRetryConfigs();
         }
 
         auto encodedServerHello = encodeHandshake(std::move(serverHello));
@@ -1362,6 +1502,7 @@ EventHandler<ServerTypes, StateEnum::ExpectingClientHello, Event::ClientHello>::
             *handshakeContext,
             alpn,
             earlyDataType,
+            std::move(echRetryConfigs),
             std::move(additionalExtensions));
 
         /*
@@ -1421,10 +1562,6 @@ EventHandler<ServerTypes, StateEnum::ExpectingClientHello, Event::ClientHello>::
 
         auto clientRandom = std::move(chlo.random);
 
-        ECHStatus echStatus = ECHStatus::NotRequested;
-        if (requestedECH) {
-          echStatus = acceptedECH ? ECHStatus::Accepted : ECHStatus::Rejected;
-        }
         return runOnCallerIfComplete(
             state.executor(),
             std::move(signature),
@@ -1610,6 +1747,7 @@ EventHandler<ServerTypes, StateEnum::ExpectingClientHello, Event::ClientHello>::
                     newState.handshakeTime() = std::move(handshakeTime);
                     newState.clientRandom() = std::move(clientRandom);
                     newState.echStatus() = echStatus;
+                    newState.echState() = folly::none;
                   });
 
               if (earlyDataType == EarlyDataType::Accepted) {
@@ -1893,10 +2031,18 @@ AsyncActions EventHandler<
       state.handshakeContext()->getHandshakeContext()->coalesce(),
       certVerify.signature->coalesce());
 
+  std::shared_ptr<const Cert> newCert;
+
   try {
     const auto& verifier = state.context()->getClientCertVerifier();
     if (verifier) {
-      verifier->verify(certs);
+      if (auto verifiedCert = verifier->verify(certs)) {
+        newCert = std::move(verifiedCert);
+      } else {
+        newCert = std::move(leafCert);
+      }
+    } else {
+      newCert = std::move(leafCert);
     }
   } catch (const FizzException&) {
     throw;
@@ -1909,7 +2055,7 @@ AsyncActions EventHandler<
   state.handshakeContext()->appendToTranscript(*certVerify.originalEncoding);
 
   return actions(
-      MutateState([cert = std::move(leafCert)](State& newState) {
+      MutateState([cert = std::move(newCert)](State& newState) {
         newState.unverifiedCertChain() = folly::none;
         newState.clientCert() = std::move(cert);
       }),
@@ -2040,6 +2186,44 @@ EventHandler<ServerTypes, StateEnum::AcceptingData, Event::AppWrite>::handle(
   return actions(std::move(write));
 }
 
+AsyncActions EventHandler<
+    ServerTypes,
+    StateEnum::AcceptingData,
+    Event::KeyUpdateInitiation>::handle(const State& state, Param param) {
+  if (state.readRecordLayer()->hasUnparsedHandshakeData()) {
+    throw FizzException("data after key_update", folly::none);
+  }
+  auto& keyUpdateInitiation = *param.asKeyUpdateInitiation();
+  auto encodedKeyUpdated =
+      Protocol::getKeyUpdated(keyUpdateInitiation.request_update);
+  WriteToSocket write;
+  write.contents.emplace_back(
+      state.writeRecordLayer()->writeHandshake(std::move(encodedKeyUpdated)));
+
+  state.keyScheduler()->serverKeyUpdate();
+
+  auto writeRecordLayer =
+      state.context()->getFactory()->makeEncryptedWriteRecordLayer(
+          EncryptionLevel::AppTraffic);
+  writeRecordLayer->setProtocolVersion(*state.version());
+  auto writeSecret =
+      state.keyScheduler()->getSecret(AppTrafficSecrets::ServerAppTraffic);
+  Protocol::setAead(
+      *writeRecordLayer,
+      *state.cipher(),
+      folly::range(writeSecret.secret),
+      *state.context()->getFactory(),
+      *state.keyScheduler());
+
+  return actions(
+      MutateState([wRecordLayer =
+                       std::move(writeRecordLayer)](State& newState) mutable {
+        newState.writeRecordLayer() = std::move(wRecordLayer);
+      }),
+      SecretAvailable(std::move(writeSecret)),
+      std::move(write));
+}
+
 AsyncActions
 EventHandler<ServerTypes, StateEnum::AcceptingData, Event::KeyUpdate>::handle(
     const State& state,
@@ -2049,6 +2233,7 @@ EventHandler<ServerTypes, StateEnum::AcceptingData, Event::KeyUpdate>::handle(
   if (state.readRecordLayer()->hasUnparsedHandshakeData()) {
     throw FizzException("data after key_update", folly::none);
   }
+
   state.keyScheduler()->clientKeyUpdate();
   auto readRecordLayer =
       state.context()->getFactory()->makeEncryptedReadRecordLayer(
