@@ -13,6 +13,7 @@
 #include <fizz/protocol/Certificate.h>
 #include <fizz/protocol/KeyScheduler.h>
 #include <fizz/protocol/Types.h>
+#include <fizz/protocol/ech/Types.h>
 #include <fizz/record/Extensions.h>
 #include <fizz/record/RecordLayer.h>
 #include <fizz/server/Actions.h>
@@ -39,6 +40,15 @@ enum class StateEnum {
 
 enum class ECHStatus { NotRequested, Accepted, Rejected };
 
+struct ECHState {
+  // Cipher suite used (for HRR)
+  ech::ECHCipherSuite cipherSuite;
+  // Config ID used (for HRR)
+  Buf configId;
+  // HPKE context saved for use with HRR, if needed.
+  mutable std::unique_ptr<hpke::HpkeContext> hpkeContext;
+};
+
 struct HandshakeLogging {
   folly::Optional<ProtocolVersion> clientLegacyVersion;
   std::vector<ProtocolVersion> clientSupportedVersions;
@@ -313,6 +323,10 @@ class State {
     return echStatus_;
   }
 
+  const folly::Optional<ECHState>& echState() const {
+    return echState_;
+  }
+
   /*
    * State setters.
    */
@@ -420,6 +434,10 @@ class State {
     return echStatus_;
   }
 
+  auto& echState() {
+    return echState_;
+  }
+
  private:
   StateEnum state_{StateEnum::Uninitialized};
 
@@ -462,6 +480,7 @@ class State {
   std::vector<uint8_t> resumptionMasterSecret_;
   folly::Optional<std::chrono::system_clock::time_point> handshakeTime_;
   ECHStatus echStatus_{ECHStatus::NotRequested};
+  folly::Optional<ECHState> echState_;
 
   std::unique_ptr<HandshakeLogging> handshakeLogging_;
 
